use crate::geom::{Line2d, Vec2, Vec3};
use crate::mat::Material;
use crate::real::Real;
use crate::sim::CellState;

use serde::{Serialize, Deserialize};

/// M(B) parallelogram
///
///```text
///     ____________
///    /           /
///   /     .     /
///  /           /
/// /___________/
/// ```
///
/// The `.` depicts (0, 0). X axis is B; y axis is M.
/// As B increases, M remains constant until it hits an edge.
/// Then M rises up to its max.
/// Same thing happens on the left edge, as B decreases and M falls to its min.
#[derive(Default, Copy, Clone, PartialEq, Serialize, Deserialize)]
pub struct MBPgram<R> {
    /// Vertical range of the graph
    pub max_m: R,
    /// X coordinate at which the upward slope starts
    pub b_start: R,
    /// X coordinate at which the upward slope ends
    pub b_end: R,
}

impl<R: Real> MBPgram<R> {
    pub fn new(b_start: R, b_end: R, max_m: R) -> Self {
        Self { b_start, b_end, max_m }
    }

    /// Return the new `M`
    pub fn move_b(&self, m: R, target_b: R) -> R {
        let right_edge = Line2d::new(
            Vec2::new(self.b_start, -self.max_m),
            Vec2::new(self.b_end, self.max_m),
        );
        let left_edge = Line2d::new(
            Vec2::new(-self.b_start, self.max_m),
            Vec2::new(-self.b_end, -self.max_m),
        );
        // m must be at least this much:
        let min_m = right_edge.clamp_by_x(target_b).y();
        // m must be no more than this:
        let max_m = left_edge.clamp_by_x(target_b).y();
        m.max_or_undefined(min_m).min_or_undefined(max_m)
    }
}

#[derive(Default, Copy, Clone, PartialEq, Serialize, Deserialize)]
pub struct NativeMBFerromagnet<R> {
    m: Vec3<R>,
    curve: MBPgram<R>,
}

impl<R: Real> NativeMBFerromagnet<R> {
    pub fn new<R2: Real>(b_start: R2, b_end: R2, max_m: R2) -> Self {
        Self {
            m: Vec3::zero(),
            curve: MBPgram::new(b_start.cast(), b_end.cast(), max_m.cast()),
        }
    }

    pub fn curve(&self) -> MBPgram<R> {
        self.curve
    }
}

impl<R: Real> Material<R> for NativeMBFerromagnet<R> {
    fn step_b(&mut self, context: &CellState<R>, delta_b: Vec3<R>) {
        let target_b = context.with_m(self.m).b() + delta_b;
        // println!("step_b {}", target_b);
        self.m = Vec3::new(
            self.curve.move_b(self.m.x(), target_b.x()),
            self.curve.move_b(self.m.y(), target_b.y()),
            self.curve.move_b(self.m.z(), target_b.z()),
        );
    }
    fn m(&self) -> Vec3<R> {
        self.m
    }
}

#[derive(Default, Copy, Clone, PartialEq, Serialize, Deserialize)]
pub struct SpirvMBFerromagnet<R>(NativeMBFerromagnet<R>);

impl<R: Real> SpirvMBFerromagnet<R> {
    pub fn new<R2: Real>(b_start: R2, b_end: R2, max_m: R2) -> Self {
        Self(NativeMBFerromagnet::new(b_start, b_end, max_m))
    }

    pub fn curve(&self) -> MBPgram<R> {
        self.0.curve()
    }
}

impl<R: Real> Material<R> for SpirvMBFerromagnet<R> {
    fn step_b(&mut self, context: &CellState<R>, delta_b: Vec3<R>) {
        let target_b = context.with_m(self.m()).b() + delta_b;
        let curve = coremem_types::mat::MBPgram::new(
            self.0.curve.b_start,
            self.0.curve.b_end,
            self.0.curve.max_m,
        );
        // println!("step_b {}", target_b);
        self.0.m = Vec3::new(
            curve.move_b(self.0.m.x(), target_b.x()),
            curve.move_b(self.0.m.y(), target_b.y()),
            curve.move_b(self.0.m.z(), target_b.z()),
        );
    }
    fn m(&self) -> Vec3<R> {
        self.0.m()
    }
}

// XXX: for debugging, use the same MBFerromagnet impl as we do in spirv impl.
// pub type MBFerromagnet<R> = SpirvMBFerromagnet<R>;
pub type MBFerromagnet<R> = NativeMBFerromagnet<R>;

#[cfg(test)]
mod test {
    use super::*;
    use float_eq::assert_float_eq;

    #[test]
    fn curve_interior() {
        let curve = MBPgram::new(4.0, 6.0, 20.0f32);

        assert_float_eq!(curve.move_b(0.0, 2.0), 0.0, abs <= 1e-5);
        assert_float_eq!(curve.move_b(0.0, 5.0), 0.0, abs <= 1e-5);
        assert_float_eq!(curve.move_b(1.0, 5.0), 1.0, abs <= 1e-5);
        assert_float_eq!(curve.move_b(20.0, 5.0), 20.0, abs <= 1e-5);
        assert_float_eq!(curve.move_b(-20.0, 4.0), -20.0, abs <= 1e-5);
        assert_float_eq!(curve.move_b(-20.0, -6.0), -20.0, abs <= 1e-5);
        assert_float_eq!(curve.move_b(20.0, -4.0), 20.0, abs <= 1e-5);
        assert_float_eq!(curve.move_b(10.0, -2.0), 10.0, abs <= 1e-5);
    }

    #[test]
    fn curve_exterior() {
        let curve = MBPgram::new(4.0, 6.0, 20.0f32);
        assert_float_eq!(curve.move_b(0.0, 6.0), 20.0, abs <= 1e-5);
        assert_float_eq!(curve.move_b(0.0, 7.0), 20.0, abs <= 1e-5);
        assert_float_eq!(curve.move_b(0.0, -6.0), -20.0, abs <= 1e-5);
        assert_float_eq!(curve.move_b(0.0, -7.0), -20.0, abs <= 1e-5);


        assert_float_eq!(curve.move_b(2.0, -6.0), -20.0, abs <= 1e-5);
        assert_float_eq!(curve.move_b(20.0, -6.0), -20.0, abs <= 1e-5);
        assert_float_eq!(curve.move_b(20.0, -5.0), 0.0, abs <= 1e-5);
        assert_float_eq!(curve.move_b(20.0, -4.5), 10.0, abs <= 1e-5);
        assert_float_eq!(curve.move_b(-15.0, 4.5), -10.0, abs <= 1e-5);
        assert_float_eq!(curve.move_b(-15.0, 5.0), 0.0, abs <= 1e-5);
        assert_float_eq!(curve.move_b(-15.0, 5.5), 10.0, abs <= 1e-5);
        assert_float_eq!(curve.move_b(-15.0, 7.5), 20.0, abs <= 1e-5);
    }

    #[test]
    fn curve_3r1() {
        // slope of 3r1 is about M=793210*B
        // This is almost identical to iron (795615)!
        let curve = MBPgram::new(-0.3899, 0.3900, 310000f32);

        // magnetizing:
        // v.s. 198893 in B(H) curve
        assert_float_eq!(curve.move_b(0.0, 0.250), 198703.0, abs <= 1.0);
        // v.s. 278321 in B(H) curve
        assert_float_eq!(curve.move_b(198703.0, 0.350), 278201.0, abs <= 1.0);
        assert_float_eq!(curve.move_b(278201.0, 0.390), 310000.0, abs <= 1.0);

        // de-magnetizing:
        // From saturation, decreasing B causes NO decrease in M: instead, it causes a decrease in
        // H. This is probably BAD: in the B(H) curve, a large change in H always causes a large
        // change in B. The movement of H here is likely to induce current, whereas it SHOULDN'T.
        assert_float_eq!(curve.move_b(310000.0, 0.38995), 310000.0, abs <= 1.0);
        // v.s. 258626 in B(H); H = 220
        assert_float_eq!(curve.move_b(310000.0, 0.325), 258406.0, abs <= 1.0);
        // here's where H crosses 0 (v.s. B=0.325 in the B(H) curve... quite a difference)
        assert_float_eq!(curve.move_b(310000.0, 0.050), 39788.438, abs <= 1.0);
        // v.s. 35.0 in B(H)
        assert_float_eq!(curve.move_b(310000.0, 0.0), 39.75, abs <= 1.0);

        // negative magnetization:
        // erase the magnetization: H = -40
        assert_float_eq!(curve.move_b(310000.0, -0.00005), 0.0, abs <= 0.1);
        // the magnetization has been completely erased:
        assert_float_eq!(curve.move_b(310000.0, -0.25), -198703.0, abs <= 1.0);
    }
}
