use serde::de::Deserializer;
use serde::ser::Serializer;
use serde::{Deserialize, Serialize};

use crate::mat::{
    AnisomorphicConductor,
    IsoConductorOr,
    IsomorphicConductor,
    Ferroxcube3R1MH,
    FullyGenericMaterial,
    MaterialExt as _,
    MBFerromagnet,
    MBPgram,
    MHPgram,
    Static
};
use crate::geom::{Index, Vec3, Vec3u};

/// hide the actual spirv backend structures inside a submodule to make their use/boundary clear.
mod ffi {
    pub use spirv_backend::entry_points;
    pub use spirv_backend::sim::SerializedSimMeta;
    pub use spirv_backend::support::Optional;
    pub use coremem_types::mat::MBPgram;
}

// conversion traits for types defined cross-lib
pub trait IntoFfi {
    type Ffi;
    fn into_ffi(self) -> Self::Ffi;
}

pub trait IntoLib {
    type Lib;
    fn into_lib(self) -> Self::Lib;
}

macro_rules! identity {
    ($($param:ident,)* => $t:ty) => {
        impl<$($param: IntoFfi),*> IntoFfi for $t {
            type Ffi = $t;
            fn into_ffi(self) -> Self::Ffi {
                self
            }
        }
        impl<$($param: IntoLib),*> IntoLib for $t {
            type Lib = $t;
            fn into_lib(self) -> Self::Lib {
                self
            }
        }
    };
}

// XXX: should work for any other lifetime, not just 'static
identity!(=> f32);
identity!(=> &'static str);
identity!(T0, T1, => (T0, T1));
identity!(=> Vec3u);
identity!(T, => Vec3<T>);

impl<L: IntoFfi> IntoFfi for Option<L>
    where L::Ffi: Default
{
    type Ffi = ffi::Optional<L::Ffi>;
    fn into_ffi(self) -> Self::Ffi {
        match self {
            Some(s) => ffi::Optional::some(s.into_ffi()),
            None => ffi::Optional::none(),
        }
    }
}

impl<F: Copy + IntoLib> IntoLib for ffi::Optional<F> {
    type Lib = Option<F::Lib>;
    fn into_lib(self) -> Self::Lib {
        if self.is_some() {
            Some(self.unwrap().into_lib())
        } else {
            None
        }
    }
}

impl IntoFfi for MBPgram<f32> {
    type Ffi = ffi::MBPgram<f32>;
    fn into_ffi(self) -> Self::Ffi {
        Self::Ffi::new(self.b_start, self.b_end, self.max_m)
    }
}

impl IntoLib for ffi::MBPgram<f32> {
    type Lib = MBPgram<f32>;
    fn into_lib(self) -> Self::Lib {
        Self::Lib::new(self.b_start, self.b_end, self.max_m)
    }
}

identity!( => MHPgram<f32>);
identity!( => Ferroxcube3R1MH);
identity!( => FullyGenericMaterial<f32>);
identity!(R, M, => IsoConductorOr<R, M>);

// N.B.: this isn't fully correct, as Static also encodes the magnetic component
impl From<Static<f32>> for FullyGenericMaterial<f32> {
    fn from(m: Static<f32>) -> Self {
        AnisomorphicConductor::new(m.conductivity).into()
    }
}

impl From<MBFerromagnet<f32>> for FullyGenericMaterial<f32> {
    fn from(m: MBFerromagnet<f32>) -> Self {
        m.curve().into_ffi().into()
    }
}

// this is bitwise- and type-compatible with the spirv SimMeta, except we need serde traits
#[derive(Clone, Default, Serialize, Deserialize)]
pub struct SimMeta {
    pub(crate) dim: Index,
    pub(crate) inv_feature_size: f32,
    pub(crate) time_step: f32,
    pub(crate) feature_size: f32,
}

impl IntoFfi for SimMeta {
    type Ffi = ffi::SerializedSimMeta;
    fn into_ffi(self) -> Self::Ffi {
        Self::Ffi {
            dim: self.dim.0.into_ffi(),
            inv_feature_size: self.inv_feature_size,
            time_step: self.time_step,
            feature_size: self.feature_size,
        }
    }
}

/// Store the FFI form in memory, but serialize via the lib form.
#[derive(Clone, Default, PartialEq)]
pub struct Remote<F>(F);

impl<F> Remote<F> {
    pub fn into_inner(self) -> F {
        self.0
    }
}

impl<L: IntoFfi> From<L> for Remote<L::Ffi> {
    fn from(l: L) -> Self {
        Remote(l.into_ffi())
    }
}

impl<F> std::ops::Deref for Remote<F> {
    type Target = F;
    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl<F> Serialize for Remote<F>
where F: Clone + IntoLib,
      F::Lib: Serialize,
{
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        let local = self.0.clone().into_lib();
        local.serialize(serializer)
    }
}

impl<'de, F> Deserialize<'de> for Remote<F>
    where F: IntoLib,
          F::Lib: Deserialize<'de> + IntoFfi<Ffi=F>,
{
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: Deserializer<'de>,
    {
        let local: F::Lib = Deserialize::deserialize(deserializer)?;
        Ok(Remote(local.into_ffi()))
    }
}

// FUNCTION BINDINGS
pub fn entry_points<L>() -> Option<(&'static str, &'static str)>
where
    L: IntoFfi,
    L::Ffi: 'static
{
    ffi::entry_points::<L::Ffi>().into_lib()
}
